"""Tools for defining a file format for storing Python information in HDF5.

This module is intended to be used declaratively: you define a file format
by specifying the attributes of a Python object, their names in HDF5, and
some other information about how they should be read and written. This
format is then an H5Format object, which can be queried for a detailed
description of the file format, suitable for inclusion in a README file.
The H5Format object also has methods to allow you to write the information
from a suitable Python object out into an HDF5 file in the described
format, and to allow you to read the information from a suitable HDF5
format into a Python object.

These formats are aimed at relatively small HDF5 files, where the data
is loaded into numpy arrays rather than being read on the fly from the
file (using HDF5 indexing).

"""
import logging
from dataclasses import dataclass
from io import StringIO
from pathlib import Path
from textwrap import dedent, indent, shorten
from typing import IO, Any, Callable, List, Optional, Union

import h5py
import numpy as np
import packaging.version

logger = logging.getLogger(__name__)


class MissingAttribute(ValueError):
    pass


class MissingName(ValueError):
    pass


@dataclass
class H5Entry:
    """An entry in an HDF5 file.

    This object describes an entry in an HDF5 file - that is,
    an instance of this class can specify that a certain attribute
    of a Python object should be written out to an HDF5 dataset
    or attribute, and how.

    There is a derived class, H5ConstantEntry, suitable for
    entries that write a particular value to the file format
    regardless of the contents of the Python object (for
    example, a format name).

    Parameters
    ----------
    name : str
        The name to be used for this entry in the HDF5 file.
        This may include spaces and need not be a valid
        Python identifier.
    description : str
        A brief description of the file format. This will
        be included in the part of the file format description
        corresponding to this entry. Indentation will be
        removed (by textwrap.dedent).
    required : bool, optional
        If True, raise an exception if this entry is not present
        in the Python object on write, or in the HDF5 file on
        read. If False, missing entries are simply not
        written.
    use_dataset : bool, optional
        If True, save the data corresponding to this entry
        in an HDF5 dataset. This is appropriate for any large
        object; the data will be compressed on write. If
        False, save the data in an attribute of the HDF5
        file. This should only be used for small items,
        often ones describing the file as a whole; strings
        in particular are more easily handled here.
    attribute : str, optional
        If supplied, this is the attribute on the Python
        object that should contain the entry. If not
        supplied, the attribute will be assumed to
        be the same as the HDF5 name.
    write : callable, optional
        If supplied, the callable is called as
        write(h5file, name, thing, attribute),
        where h5file is an open HDF5 file object.
        The callable should normally extract the
        named attribute from the provided thing
        and save it to the HDF5 file, but it may,
        for example, look up other attributes and
        save their data as attributes of a dataset.
        If this returns a dataset, a decription
        of the entry will be attached as an
        attribute.
    read : callable, optional
        If supplied, the callable is called as
        read(h5file, name, thing, attribute),
        where h5file is an open HDF5 file object.
        This callable should normally extract
        the information from the HDF5 dataset or
        attribute specified by name and save it
        as the named attribute of thing. It may
        extract auxiliary information and do
        save it in thing as well.
    extra_attributes : dict, optional
        If provided, this should be a dictionary of
        additional things to save as attributes
        of the created dataset. It is an error
        to supply this when use_dataset is False.
        Note that extra attributes supplied this
        way must be the same for all files in this
        format; if you want to store attributes
        that depend on the thing being saved,
        you will need to use read and write, above.
    """

    name: str
    description: str
    required: bool = True
    use_dataset: bool = False
    attribute: Optional[str] = None
    write: Optional[Callable] = None
    read: Optional[Callable] = None
    extra_attributes: Optional[dict] = None

    def write_to_hdf5(self, h5file: h5py.File, thing):
        """Write this entry to an HDF5 file.

        Parameters
        ----------
        h5file : h5py.File
            An open HDF5 file to write to.
        thing : object
            The relevant data is extracted from this object.
        """
        attribute = self.name if self.attribute is None else self.attribute
        if not hasattr(thing, attribute):
            if self.required:
                raise MissingAttribute(f"Attribute {attribute} needed for HDF5 {self.name} missing from {thing}")
            logger.debug(f"Not writing {self.name} because attribute {attribute} is missing")
            return
        if self.write is not None:
            n = self.write(h5file, self.name, thing, attribute)
            if n is not None:
                n.attrs["Description"] = shorten(self.description, 256)
            return
        self._write_value_to_hdf5(h5file, getattr(thing, attribute))

    def _write_value_to_hdf5(self, h5file: h5py.File, value: Any):
        if self.use_dataset:
            if isinstance(value, dict):
                new_item = write_dict_to_hdf5(h5file, self.name, value)
            elif isinstance(value, str):
                new_item = write_string_to_hdf5_dataset(h5file, self.name, value)
            else:
                new_item = write_array_to_hdf5_dataset(h5file, self.name, np.asarray(value))
            new_item.attrs["Description"] = shorten(self.description, 256)
            if self.extra_attributes is not None:
                new_item.attrs.update(self.extra_attributes)
        else:
            if isinstance(value, np.ndarray) and value.dtype.kind == "U":
                if len(value.shape) != 1:
                    raise ValueError("Cannot store multidimentsional string arrays in HDF5 attributes: {value.shape}")
                value = [str(v) for v in value]
            try:
                h5file.attrs[self.name] = value
            except TypeError as e:
                raise TypeError(f"Invalid type for storage in an attribute: {type(value)}") from e
            if self.extra_attributes is not None:
                raise ValueError(f"Cannot apply extra attributes to attribute: {self.extra_attributes}")

    def read_from_hdf5(self, h5file: h5py.File, thing):
        """Read this entry from an HDF5 file.

        Parameters
        ----------
        h5file : h5py.File
            An open HDF5 file to read the value from.
        thing : object
            Attributes of this object are set to the values extracted
            from the HDF5 file.        logger.debug(f"Creating non-string dataset {name}")

        """
        attribute = self.name if self.attribute is None else self.attribute
        if self.name not in (h5file if self.use_dataset else h5file.attrs):
            if self.required:
                raise MissingName(f"Entry {self.name} missing from HDF5")
            logger.debug(f"Not reading {attribute} because attribute {self.name} is missing")
            return
        if self.read is not None:
            return self.read(h5file, self.name, thing, attribute)
        try:
            if self.use_dataset:
                value = h5file[self.name]
                if isinstance(value, h5py.Group):
                    value = read_dict_from_hdf5(value)
                else:
                    value = decode_array_dataset_if_necessary(value)
            else:
                value = h5file.attrs[self.name]
                if isinstance(value, np.ndarray) and value.dtype == "O" and len(value.shape) == 1:
                    value = list(value)
        except KeyError:
            if self.required:
                raise
            return
        else:
            setattr(thing, attribute, value)

    def write_description(self, f: IO[str], extra_tags: Optional[List[str]] = None):
        """Write out a description of this entry.

        This will give the name and properties of this entry in
        a format suitable for inclusion in the overall description
        of the file format. In particular each entry produces one
        item in a Markdown list.

        Parameters
        ----------
        f : file
            An open file to write the description to.
        extra_tags : list of str, optional
            If provided, each item in the list will be
            added to the list of tags on the first line
            of the description.
        """
        tags = ["dataset" if self.use_dataset else "attribute"]
        if not self.required:
            tags.append("optional")
        if self.extra_attributes is not None and "units" in self.extra_attributes:
            tags.append(f"units={self.extra_attributes['units']}")
        if extra_tags is not None:
            tags.extend(extra_tags)
        # End with two spaces to arrange for a Markdown line break
        print(f"* `{self.name}` ({', '.join(tags)})  ", file=f)
        print(indent(dedent(self.description).strip(), 4 * " "), file=f)


@dataclass
class H5ConstantEntry(H5Entry):
    """An H5Entry that always stores the same value.

    This kind of entry is constructed with a constant value
    that is stored in every file in this format. This
    is suitable for format names, version numbers, or
    other metadata about the format itself.

    Upon read, the loaded data is stored into the indicated
    attribute of the Python object, and not checked against
    the constant value; users who want to do validation
    should check the values themselves.
    """

    value: Any = None

    def write_to_hdf5(self, h5file: h5py.File, thing):
        self._write_value_to_hdf5(h5file, self.value)

    def write_description(self, f: IO[str], extra_tags: Optional[List[str]] = None):
        tags = [f"constant value={repr(self.value)}"]
        if extra_tags is not None:
            tags.extend(extra_tags)
        super().write_description(f, extra_tags=tags)


def write_array_to_hdf5_dataset(h5group: h5py.Group, name: str, value: np.ndarray) -> h5py.Dataset:
    """Write an array to an HDF5 dataset.

    For numeric arrays, this just creates a dataset and saves the value.

    For strings, the string is split at newline characters, the lines
    are encoded in UTF-8, and the result is written as a (fixed-length)
    byte array. An attribute "lines" is set to True to indicate that
    this contains the lines of a single string. Attributes
    "coding" and "type" are set to "utf-8" and "str' respectively.

    For arrays of strings, the strings are converted to UTF-8 (but not
    further split) and stored as a fixed-length byte array. Attributes
    "coding" and "type" are set to "utf-8" and "str' respectively.

    The dataset is constructed with gzip compression and the shuffle filter,
    unless the value is a simple scalar.
    """
    if value.dtype.kind == "U":
        logger.debug(f"Encoding {name} {value} as dataset in utf-8")
        encoded = True
        value = np.char.encode(value, "utf-8")
    else:
        encoded = False
    if value.shape == ():
        logger.debug(f"Creating string dataset {name}")
        d = h5group.create_dataset(
            name,
            data=value,
            track_order=True,
        )
    else:
        logger.debug(f"Creating non-string dataset {name}")
        d = h5group.create_dataset(
            name,
            data=value,
            compression="gzip",
            compression_opts=9,
            shuffle=True,
            track_order=True,
        )
    if encoded:
        logger.debug(f"Recording attributes for {name}")
        d.attrs["lines"] = False
        d.attrs["coding"] = "utf-8"
        d.attrs["type"] = "str"
    return d


def write_string_to_hdf5_dataset(h5group: h5py.Group, name: str, value: str) -> h5py.Dataset:
    """Write a string to an HDF5 dataset.

    This is for writing single, long, Python strings; it splits them at
    newlines and then saves them as a byte array.
    """
    value_as_array = np.array([s.encode("utf-8") for s in value.split("\n")])
    logger.debug(f"converted {repr(value)} to {repr(value_as_array)}")
    d = write_array_to_hdf5_dataset(h5group, name, value_as_array)
    d.attrs["lines"] = True
    d.attrs["coding"] = "utf-8"
    d.attrs["type"] = "str"
    return d


def decode_array_dataset_if_necessary(dataset: h5py.Dataset) -> Any:
    """Extract the value from a dataset.

    This exists to extract values from an HDF5 dataset, reversing any
    encoding or line splitting that has been done. It relies on the
    attributes "lines", "type", and "coding" of the dataset to
    determine what decoding is appropriate.
    """
    lines = dataset.attrs.get("lines", False)
    type_ = dataset.attrs.get("type", "")
    coding = dataset.attrs.get("coding", "")

    logger.debug(f"Decoding {dataset} with type={type_} coding={coding} lines={lines}")
    if type_ == "str":
        if lines:
            return "\n".join([s.decode(coding) for s in dataset])
        return np.char.decode(dataset, coding)
    return np.array(dataset)


def write_dict_to_hdf5(h5group: h5py.Group, name: str, d: dict) -> h5py.Group:
    """Write a dictionary to an HDF5 file (or group).

    Because HDF5 doesn't have a native notion of a dictionary,
    this uses HDF5 groups to store the data. Thus the keys
    must be strings. Contents of the dictionary are stored
    as attributes if scalars, datasets if arrays, and sub-groups
    if they are dicts.
    """
    g = h5group.create_group(name, track_order=True)
    for k, v in d.items():
        if k == "Description":
            raise ValueError("The value `Description` is special to h5format and cannot be stored in dictionaries.")
        if isinstance(v, dict):
            write_dict_to_hdf5(g, k, v)
        elif isinstance(v, np.ndarray):
            write_array_to_hdf5_dataset(g, k, v)
        else:
            g.attrs[k] = v
    return g


def read_dict_from_hdf5(h5group: h5py.Group) -> dict:
    """Read a dictionary from an HDF5 file (or group).

    This converts HDF5 groups into dictionaries, saving
    their attributes and datasets as values in the dictionary,
    and sub-groups as further dictionaries.
    """
    r = dict(h5group.attrs)
    r.pop("Description", None)
    for k, v in h5group.items():
        r[k] = read_dict_from_hdf5(v) if isinstance(v, h5py.Group) else decode_array_dataset_if_necessary(v)
    return r


class H5Format:
    """An HDF5-based file format.

    These objects exist to describe, write, and read HDF5-based
    file formats. Normally an H5Format is constructed by a
    library, and users can use this object to read and write
    library objects; they can also obtain a Markdown description
    of the format by querying the object.

    Parameters
    ----------
    description_intro : str
        Some text to supply as an introduction. This will be
        run through textwrap.dedent to clean up leading indentation.
    entries : list of H5Entry
        The entries in the file format.
    description_finale : str, optional
        A final bit of text to add to the description.
    format_name : str, optional
        A distinctive name for this file format. This name will
        be checked against files read and a warning will be issued
        if there is a mismatch.
    format_version : str, optional
        A version string defining the current version of the file
        format. This version string should be of the form 1.2.3,
        as a form of semantic versioning. A warning will be issued
        if a file being read is not guaranteed to be readable
        by the rules of semantic versioning (major versions are
        not guaranteed any compatibility, newer minor versions
        should be able to read older minor versions but not
        necessarily vice versa).
    """

    def __init__(
        self,
        description_intro: str,
        entries: Optional[List[H5Entry]] = None,
        description_finale: str = "",
        format_name: Optional[str] = None,
        format_version: Optional[str] = None,
    ):
        self.description_intro = dedent(description_intro)
        self.description_finale = dedent(description_finale)
        self.format_name = format_name
        self.format_version = format_version
        self.entries: List[H5Entry] = []
        if self.format_name is not None:
            self.entries.append(
                H5ConstantEntry(
                    name="format_name",
                    required=False,
                    description="The name of this particular HDF5 format.",
                    value=self.format_name,
                )
            )
        if self.format_version is not None:
            if not isinstance(packaging.version.parse(self.format_version), packaging.version.Version):
                raise ValueError(f"Unable to parse format_version {self.format_version}")
            self.entries.append(
                H5ConstantEntry(
                    name="format_version",
                    required=False,
                    description="""\
                        Version number indicating the compatibility of this file with
                        other readers of this format.
                        """,
                    value=self.format_version,
                )
            )
        if entries is not None:
            self.entries.extend(entries)

    def add_entry(self, entry: H5Entry):
        """Add a new entry to the end of the list in the format."""
        logger.debug(f"Added entry {entry.name} ({entry.attribute})")
        self.entries.append(entry)

    def write_description(self, f: IO[str]):
        """Write the format description to an open file."""
        f.write(self.description_intro)
        for entry in self.entries:
            entry.write_description(f)
        f.write("\n")
        f.write(self.description_finale)

    def save_to_hdf5(self, h5: Union[Path, str, h5py.Group], thing):
        """Save the information in thing to HDF5.

        Parameters
        ----------
        h5 : str or Path or h5py.File or h5py.Group
            Where to write the file. If a string or path, that is opened
            as an HDF5 file for writing; otherwise an open HDF5 file
            should be provided. It is possible to write to a sub-group
            of an existing file, although the reader will need to
            be provided the appropriate sub-group to work with.
        thing : object
            A python object containing the data to be written out,
            extracting data from its attributes as specified by the
            entries in this file format.
        """
        if not isinstance(h5, h5py.Group):
            # This requests the the file preserve insertion order;
            # it may not work on read: https://github.com/h5py/h5py/issues/1577
            with h5py.File(h5, "w", track_order=True) as f:
                return self.save_to_hdf5(f, thing)
        H5ConstantEntry(
            name="README", description="A text description of the format.", use_dataset=True, value=self.description
        ).write_to_hdf5(h5, thing)
        for entry in self.entries:
            entry.write_to_hdf5(h5, thing)

    def load_from_hdf5(self, h5: Union[Path, str, h5py.Group], thing):
        """Load data from an HDF5 file into an existing Python object.

        This method will emit warnings if the file does not appear to
        be in a compatible format, or raise an exception if it cannot
        be read as an HDF5 file at all.

        Parameters
        ----------
        h5 : str or Path or h5py.File or h5py.Group
            Where to read the data. If a string or path, that is opened
            as an HDF5 file for reading; otherwise an open HDF5 file
            should be provided. It is possible to read from a sub-group
            of an existing file.
        thing : object
            A Python object to write the data to. Note that the Python
            type "object" is unsuitable as it cannot accept additional
            attributes, but most Python objects are appropriate here.
        """
        if not isinstance(h5, h5py.Group):
            with h5py.File(h5, "r") as f:
                return self.load_from_hdf5(f, thing)
        for entry in self.entries:
            entry.read_from_hdf5(h5, thing)
        if self.format_name is not None and getattr(thing, "format_name") != self.format_name:
            logger.warning(
                f"Apparently different formats: file is in format "
                f"{getattr(thing, 'format_name')} while this reader expects "
                f"format {self.format_name}."
            )
        if self.format_version is not None and hasattr(thing, "format_version"):
            s = packaging.version.parse(self.format_version)
            o = packaging.version.parse(thing.format_version)
            if s.major != o.major:
                logger.warning(
                    f"Incompatible major versions for the format ({self.format_version} and {thing.format_version})"
                )
            if s.minor < o.minor:
                logger.warning(
                    f"File has newer format minor version than reader "
                    f"({self.format_version} and {thing.format_version})"
                )

    @property
    def description(self) -> str:
        """A (long) string describing the file format."""
        f = StringIO()
        self.write_description(f)
        return f.getvalue()

    @property
    def all_names(self) -> List[str]:
        """All names that will be written to the HDF5 file.

        This can be used to look for other information, not used by this format,
        that has been stored in the HDF5 file.
        """
        return ["README"] + [e.name for e in self.entries]



# pulsar.py
"""Class containing pulsar data from timing package [tempo2/PINT].
"""

import contextlib
import json
import logging
import os
import pickle
from io import StringIO
from typing import Optional

import astropy.constants as const
import astropy.units as u
import numpy as np
from ephem import Ecliptic, Equatorial

import enterprise
from enterprise.signals import utils
from enterprise.pulsar_inflate import PulsarInflater

from h5pulsar.h5format import H5Format
from h5pulsar import derivative_file

logger = logging.getLogger(__name__)

try:
    import libstempo as t2
except (ImportError, RuntimeError) as e:
    # ImportError happens if libstempo isn't installed
    # RuntimeError happens if libstempo is installed but can't find TEMPO2
    logger.warning(f"Unable to use TEMPO2: {e}. Will use PINT instead.")
    t2 = None

try:
    import pint
    from pint.models import TimingModel, get_model_and_toas
    from pint.residuals import Residuals as resids
    from pint.toa import TOAs
except ImportError:
    logger.warning("PINT not installed. Will use libstempo instead.")
    pint = None

# FIXME: maybe not needed if we have HDF5?
if pint is None and t2 is None:
    raise ImportError("Must have either PINT or libstempo timing package installed")

# Causes a circular import
# try:
#     import enterprise.derivative_file
# except ImportError as e:
#     logger.warning(f"HDF5 library not available; consider installing h5py ({e})")
#     enterprise.derivative_file = None


def get_maxobs(timfile):
    """Utility function to return number of lines in tim file.
    :param timfile:
        Full path to tim-file. For tim-files that use INCLUDEs this
        should be the base tim file.
    :returns: Number of lines in tim-file
    """

    maxobs = 0
    with open(timfile) as tfile:
        flines = tfile.readlines()
        lines = [ln for ln in flines if not ln.startswith("C")]
        if any(["INCLUDE" in ln for ln in lines]):
            for line in [ln for ln in lines if "INCLUDE" in ln]:
                maxobs += get_maxobs(line.split()[-1])
        else:
            maxobs = sum(1 for line in lines if line.rstrip("\n"))
    return maxobs


class BasePulsar(object):
    """Abstract Base Class for Pulsar objects."""

    def _get_pdist(self):
        dfile = enterprise.__path__[0] + "/datafiles/pulsar_distances.json"
        with open(dfile, "r") as fl:
            pdict = json.load(fl)

        if self.name[0] not in ["J", "B"]:
            if "J" + self.name in pdict:
                pdist = tuple(pdict.get("J" + self.name))
            else:
                pdist = tuple(pdict.get("B" + self.name, (1.0, 0.2)))
        else:
            pdist = tuple(pdict.get(self.name, (1.0, 0.2)))

        if pdist == (1.0, 0.2):
            msg = "WARNING: Could not find pulsar distance for "
            msg += "PSR {0}.".format(self.name)
            msg += " Setting value to 1 with 20% uncertainty."
            logger.warning(msg)
        return pdist

    def _get_radec_from_ecliptic(self, elong, elat):
        # convert via pyephem
        try:
            ec = Ecliptic(elong, elat)

            # check for B name
            if "B" in self.name:
                epoch = "1950"
            else:
                epoch = "2000"
            eq = Equatorial(ec, epoch=str(epoch))
            raj = np.double(eq.ra)
            decj = np.double(eq.dec)

        except TypeError:
            msg = "WARNING: Cannot find sky location coordinates "
            msg += "for PSR {0}. ".format(self.name)
            msg += "Setting values to 0.0"
            logger.warning(msg)
            raj = 0.0
            decj = 0.0

        return raj, decj

    def _get_pos(self):
        return np.array(
            [np.cos(self._raj) * np.cos(self._decj), np.sin(self._raj) * np.cos(self._decj), np.sin(self._decj)]
        )

    def sort_data(self):
        """Sort data by time."""
        if self._sort:
            self._isort = np.argsort(self._toas, kind="mergesort")
            self._iisort = np.zeros(len(self._isort), dtype=int)
            for ii, p in enumerate(self._isort):
                self._iisort[p] = ii
        else:
            self._isort = slice(None, None, None)
            self._iisort = slice(None, None, None)

    def filter_data(self, start_time=None, end_time=None):
        """Filter data to create a time-slice of overall dataset."""
        if start_time is None and end_time is None:
            mask = np.ones(self._toas.shape, dtype=bool)
        else:
            mask = np.logical_and(self._toas >= start_time * 86400, self._toas <= end_time * 86400)

        self._toas = self._toas[mask]
        self._toaerrs = self._toaerrs[mask]
        self._residuals = self._residuals[mask]
        self._ssbfreqs = self._ssbfreqs[mask]

        self._designmatrix = self._designmatrix[mask, :]
        dmx_mask = np.sum(self._designmatrix, axis=0) != 0.0
        self._designmatrix = self._designmatrix[:, dmx_mask]

        if isinstance(self._flags, np.ndarray):
            self._flags = self._flags[mask]
        else:
            for key in self._flags:
                self._flags[key] = self._flags[key][mask]

        if self._planetssb is not None:
            self._planetssb = self.planetssb[mask, :, :]

        self.sort_data()

    def drop_not_picklable(self):
        """Drop all attributes that cannot be pickled.

        Derived classes should implement this if they have
        any such attributes.
        """
        pass

    def to_pickle(self, outdir=None):
        """Save object to pickle file."""

        self.drop_not_picklable()

        if outdir is None:
            outdir = os.getcwd()

        if not os.path.exists(outdir):
            os.makedirs(outdir)

        with open(outdir + "/{0}.pkl".format(self.name), "wb") as f:
            pickle.dump(self, f)

    @property
    def isort(self):
        """Return sorting indices."""
        return self._isort

    @property
    def iisort(self):
        """Return inverse of sorting indices."""
        return self._iisort

    @property
    def toas(self):
        """Return array of TOAs in seconds."""
        return self._toas[self._isort]

    @property
    def stoas(self):
        """Return array of observatory TOAs in seconds."""
        return self._stoas[self._isort]

    @property
    def residuals(self):
        """Return array of residuals in seconds."""
        return self._residuals[self._isort]

    @property
    def toaerrs(self):
        """Return array of TOA errors in seconds."""
        return self._toaerrs[self._isort]

    @property
    def freqs(self):
        """Return array of radio frequencies in MHz."""
        return self._ssbfreqs[self._isort]

    @property
    def Mmat(self):
        """Return ntoa x npar design matrix."""
        return self._designmatrix[self._isort, :]

    @property
    def pdist(self):
        """Return tuple of pulsar distance and uncertainty in kpc."""
        return self._pdist

    @property
    def dm(self):
        """Return DM parameter from parfile."""
        return self._dm

    @property
    def dmx(self):
        """Return a dictionary of DMX-parameter values and stoa ranges
        from parfile."""
        return self._dmx

    @property
    def flags(self):
        """Return a dictionary of tim-file flags."""

        flagnames = self._flags.dtype.names if isinstance(self._flags, np.ndarray) else self._flags.keys()

        return {flag: self._flags[flag][self._isort] for flag in flagnames}

    def set_flags(self, flagname, values):
        """Set value of existing or new flags."""

        if isinstance(self._flags, np.ndarray):
            raise NotImplementedError("Cannot set flags when stored as numpy.ndarray.")
        else:
            self._flags[flagname] = values[self._iisort]

    @property
    def backend_flags(self):
        """Return array of backend flags.

        Not all TOAs have the same flags for all data sets. In order to
        facilitate this we have a ranked ordering system that will look
        for flags. The order is `group`, `g`, `sys`, `i`, `f`, `fe`+`be`.

        """

        # collect flag names
        flagnames = self._flags.dtype.names if isinstance(self._flags, np.ndarray) else list(self._flags.keys())

        # allocate array with widest dtype
        ret = np.zeros(len(self._toas), dtype=max([self._flags[name].dtype for name in flagnames]))

        # go through the flags in reverse order of preference
        # setting or replacing values for each TOA

        if "fe" in flagnames and "be" in flagnames:
            ret[:] = [(a + "_" + b if (a and b) else "") for a, b in zip(self._flags["fe"], self._flags["be"])]

        for flag in ["f", "i", "sys", "g", "group"]:
            if flag in flagnames:
                ret[:] = np.where(self._flags[flag] == "", ret, self._flags[flag])

        return ret[self._isort]

    @property
    def theta(self):
        """Return polar angle of pulsar in radians."""
        return np.pi / 2 - self._decj

    @property
    def phi(self):
        """Return azimuthal angle of pulsar in radians."""
        return self._raj

    @property
    def pos(self):
        """Return unit vector from SSB to pulsar at fiducial POSEPOCH."""
        return self._pos

    @property
    def pos_t(self):
        """Return unit vector from SSB to pulsar as function of time."""
        return self._pos_t[self._isort, :]

    @property
    def planetssb(self):
        """Return planetary position vectors at all timestamps"""
        return self._planetssb[self._isort, :, :]

    @property
    def sunssb(self):
        """Return sun position vector at all timestamps"""
        return self._sunssb[self._isort, :]

    @property
    def telescope(self):
        """Return telescope name at all timestamps"""
        return self._telescope[self._isort]

    def to_hdf5(self, h5path, fmt=None):
        """Save this object to an HDF5 format."""
        if fmt is None:
            fmt = derivative_file.derivative_format()
        # Save MJDI, MJDF to retain full accuracy without longdouble
        # FIXME: can T2Pulsar provide this?
        if hasattr(self, "pint_toas"):
            logger.info("Creating MJDI/MJDF parameters")
            tdbs = [t.tdb for t in self.pint_toas["mjd"]]
            jd1, jd2 = np.array([(t.jd1, t.jd2) for t in tdbs]).T
            mjd1 = jd1 - 2400000
            mjd2 = jd2 - 0.5
            mjd1 += mjd2 // 1
            mjd2 = mjd2 % 1
            self.mjdi = mjd1.astype(np.int64)
            self.mjdf = mjd2
        fmt.save_to_hdf5(h5path, self)
        # FIXME: try/finally to remove mjdi/mjdf?


class PintPulsar(BasePulsar):
    def __init__(self, toas, model, sort=True, drop_pintpsr=True, planets=True):

        self._sort = sort
        self.planets = planets
        self.name = model.PSR.value

        if not drop_pintpsr:
            self.model = model
            self.parfile = model.as_parfile()
            self.pint_toas = toas
            with StringIO() as tim:
                toas.write_TOA_file(tim)
                self.timfile = tim.getvalue()

        # these are TDB but not barycentered
        # self._toas = np.array(toas.table["tdbld"], dtype="float64") * 86400
        self._toas = np.array(model.get_barycentric_toas(toas).value, dtype="float64") * 86400
        # saving also stoas (e.g., for DMX comparisons)
        self._stoas = np.array(toas.get_mjds().value, dtype="float64") * 86400
        self._residuals = np.array(resids(toas, model).time_resids.to(u.s), dtype="float64")
        self._toaerrs = np.array(toas.get_errors().to(u.s), dtype="float64")
        self._designmatrix, self.fitpars, self.designmatrix_units = model.designmatrix(toas)
        self._ssbfreqs = np.array(model.barycentric_radio_freq(toas), dtype="float64")
        self._telescope = np.array(toas.get_obss())

        # gather DM/DMX information if available
        self._set_dm(model)

        # set parameters
        self.setpars = [sp for sp in model.params if sp not in self.fitpars]

        # FIXME: this can be done more cleanly using PINT
        self._flags = {}
        for ii, obsflags in enumerate(toas.get_flags()):
            for jj, flag in enumerate(obsflags):
                if flag not in list(self._flags.keys()):
                    self._flags[flag] = [""] * toas.ntoas

                self._flags[flag][ii] = obsflags[flag]

        # convert flags to arrays
        # TODO probably better way to do this
        # in fact PINT always stores flags as strings
        for key, val in self._flags.items():
            if isinstance(val[0], u.quantity.Quantity):
                self._flags[key] = np.array([v.value for v in val])
            else:
                self._flags[key] = np.array(val)

        self._pdist = self._get_pdist()
        self._raj, self._decj = self._get_radec(model)
        self._pos = self._get_pos()
        self._planetssb = self._get_planetssb(toas, model)
        self._sunssb = self._get_sunssb(toas, model)

        which_astrometry = (
            "AstrometryEquatorial" if "AstrometryEquatorial" in model.components else "AstrometryEcliptic"
        )

        self._pos_t = model.components[which_astrometry].ssb_to_psb_xyz_ICRS(model.get_barycentric_toas(toas)).value

        self.sort_data()

    def drop_pintpsr(self):
        with contextlib.suppress(NameError):
            del self.model
            del self.parfile
            del self.pint_toas
            del self.timfile

    def drop_not_picklable(self):
        with contextlib.suppress(AttributeError):
            del self.model
            del self.pint_toas
            logger.warning("pint_toas and model objects cannot be pickled and have been removed.")

        return super().drop_not_picklable()

    def _set_dm(self, model):
        pars = [par for par in model.params if not getattr(model, par).frozen]

        if hasattr(model, "DM"):
            self._dm = float(model["DM"].value)

        dmx = {
            par: {
                "DMX": float(model[par].value),
                "DMXerr": None if model[par].uncertainty_value is None else float(model[par].uncertainty_value),
                "DMXR1": float(model[par[:3] + "R1" + par[3:]].value),
                "DMXR2": float(model[par[:3] + "R2" + par[3:]].value),
                "fit": par in pars,
            }
            for par in pars
            if "DMX_" in par
        }

        if dmx:
            self._dmx = dmx

    def _get_radec(self, model):
        if hasattr(model, "RAJ") and hasattr(model, "DECJ"):
            return (model.RAJ.value, model.DECJ.value)
        else:
            # TODO: better way of dealing with units
            d2r = np.pi / 180
            elong, elat = model.ELONG.value, model.ELAT.value
            return self._get_radec_from_ecliptic(elong * d2r, elat * d2r)

    def _get_ssb_lsec(self, toas, obs_planet):
        """Get the planet to SSB vector in lightseconds from Pint table"""
        if obs_planet not in toas.table.colnames:
            err_msg = f"{obs_planet} is not in toas.table.colnames. Either "
            err_msg += "`planet` flag is not True  in `toas` or further Pint "
            err_msg += "development to add additional planets is needed."
            raise ValueError(err_msg)
        vec = toas.table[obs_planet] + toas.table["ssb_obs_pos"]
        return (vec / const.c).to("s").value

    def _get_planetssb(self, toas, model):
        planetssb = None
        """
        Currently Pint only has position vectors for:
        [Earth, Jupiter, Saturn, Uranus, Neptune]
        No velocity vectors available
        [Mercury, Venus, Mars, Pluto] unavailable pending Pint enhancements.
        """
        if self.planets:
            planetssb = np.empty((len(self._toas), 9, 6))
            planetssb[:] = np.nan
            planetssb[:, 2, :3] = self._get_ssb_lsec(toas, "obs_earth_pos")
            planetssb[:, 4, :3] = self._get_ssb_lsec(toas, "obs_jupiter_pos")
            planetssb[:, 5, :3] = self._get_ssb_lsec(toas, "obs_saturn_pos")
            planetssb[:, 6, :3] = self._get_ssb_lsec(toas, "obs_uranus_pos")
            planetssb[:, 7, :3] = self._get_ssb_lsec(toas, "obs_neptune_pos")

            # if hasattr(model, "ELAT") and hasattr(model, "ELONG"):
            #     for ii in range(9):
            #         planetssb[:, ii, :3] = utils.ecl2eq_vec(planetssb[:, ii, :3])
            #         # planetssb[:, ii, 3:] = utils.ecl2eq_vec(planetssb[:, ii, 3:])
        return planetssb

    def _get_sunssb(self, toas, model):
        sunssb = None
        if self.planets:
            sunssb = np.zeros((len(self._toas), 6))
            sunssb[:, :3] = self._get_ssb_lsec(toas, "obs_sun_pos")

            # if hasattr(model, "ELAT") and hasattr(model, "ELONG"):
            #     sunssb[:, :3] = utils.ecl2eq_vec(sunssb[:, :3])
            # #     sunssb[:, 3:] = utils.ecl2eq_vec(sunssb[:, 3:])
        return sunssb


class Tempo2Pulsar(BasePulsar):
    def __init__(
        self,
        t2pulsar,
        sort=True,
        drop_t2pulsar=True,
        planets=True,
        par_name=None,
        tim_name=None,
    ):

        self._sort = sort
        self.t2pulsar = t2pulsar
        self.planets = planets
        self.name = str(t2pulsar.name)

        self._toas = np.double(t2pulsar.toas()) * 86400
        # saving also stoas (e.g., for DMX comparisons)
        self._stoas = np.double(t2pulsar.stoas) * 86400
        self._residuals = np.double(t2pulsar.residuals())
        self._toaerrs = np.double(t2pulsar.toaerrs) * 1e-6
        self._designmatrix = np.double(t2pulsar.designmatrix())
        self._ssbfreqs = np.double(t2pulsar.ssbfreqs()) / 1e6
        self._telescope = np.char.decode(t2pulsar.telescope(), encoding="ascii")

        # fitted parameters
        self.fitpars = ["Offset"] + [str(p) for p in t2pulsar.pars()]

        # set parameters
        spars = [str(p) for p in t2pulsar.pars(which="set")]
        self.setpars = [sp for sp in spars if sp not in self.fitpars]

        flags = {}
        for key in t2pulsar.flags():
            flags[key] = t2pulsar.flagvals(key)

        # new-style storage of flags as a numpy record array (previously, psr._flags = flags)
        self._flags = np.zeros(len(self._toas), dtype=[(key, val.dtype) for key, val in flags.items()])
        for key, val in flags.items():
            self._flags[key] = val

        self._pdist = self._get_pdist()
        self._raj, self._decj = self._get_radec(t2pulsar)
        self._pos = self._get_pos()
        self._planetssb = self._get_planetssb(t2pulsar)
        self._sunssb = self._get_sunssb(t2pulsar)

        # gather DM/DMX information if available
        self._set_dm(t2pulsar)

        self._pos_t = t2pulsar.psrPos.copy()
        if "ELONG" and "ELAT" in np.concatenate((t2pulsar.pars(which="fit"), t2pulsar.pars(which="set"))):
            self._pos_t = utils.ecl2eq_vec(self._pos_t)

        self.sort_data()

        if drop_t2pulsar:
            del self.t2pulsar
        else:
            if par_name is not None and os.path.exists(par_name):
                self.parfile = open(par_name).read()
            if tim_name is not None and os.path.exists(tim_name):
                self.timfile = open(tim_name).read()

    # gather DM/DMX information if available
    def _set_dm(self, t2pulsar):
        pars = t2pulsar.pars(which="set")

        if "DM" in pars:
            self._dm = float(t2pulsar["DM"].val)

        dmx = {
            par: {
                "DMX": float(t2pulsar[par].val),
                "DMXerr": float(t2pulsar[par].err),
                "DMXR1": float(t2pulsar[par[:3] + "R1" + par[3:]].val),
                "DMXR2": float(t2pulsar[par[:3] + "R2" + par[3:]].val),
                "fit": par in pars,
            }
            for par in pars
            if "DMX_" in par
        }

        if dmx:
            self._dmx = dmx

    def _get_radec(self, t2pulsar):
        if "RAJ" in np.concatenate((t2pulsar.pars(which="fit"), t2pulsar.pars(which="set"))):
            return (np.double(t2pulsar["RAJ"].val), np.double(t2pulsar["DECJ"].val))

        else:
            # use ecliptic coordinates
            elong = t2pulsar["ELONG"].val
            elat = t2pulsar["ELAT"].val
            return self._get_radec_from_ecliptic(elong, elat)

    def _get_planetssb(self, t2pulsar):
        planetssb = None
        if self.planets:
            for ii in range(1, 10):
                tag = "DMASSPLANET" + str(ii)
                self.t2pulsar[tag].val = 0.0
            self.t2pulsar.formbats()
            planetssb = np.zeros((len(self._toas), 9, 6))
            planetssb[:, 0, :] = self.t2pulsar.mercury_ssb
            planetssb[:, 1, :] = self.t2pulsar.venus_ssb
            planetssb[:, 2, :] = self.t2pulsar.earth_ssb
            planetssb[:, 3, :] = self.t2pulsar.mars_ssb
            planetssb[:, 4, :] = self.t2pulsar.jupiter_ssb
            planetssb[:, 5, :] = self.t2pulsar.saturn_ssb
            planetssb[:, 6, :] = self.t2pulsar.uranus_ssb
            planetssb[:, 7, :] = self.t2pulsar.neptune_ssb
            planetssb[:, 8, :] = self.t2pulsar.pluto_ssb

            if "ELONG" and "ELAT" in np.concatenate((t2pulsar.pars(), t2pulsar.pars(which="set"))):
                for ii in range(9):
                    planetssb[:, ii, :3] = utils.ecl2eq_vec(planetssb[:, ii, :3])
                    planetssb[:, ii, 3:] = utils.ecl2eq_vec(planetssb[:, ii, 3:])
        return planetssb

    def _get_sunssb(self, t2pulsar):
        sunssb = None
        if self.planets:
            # for ii in range(1, 10):
            #     tag = 'DMASSPLANET' + str(ii)@pytest.mark.skipif(t2 is None, reason="TEMPO2/libstempo not available")

            self.t2pulsar.formbats()
            sunssb = np.zeros((len(self._toas), 6))
            sunssb[:, :] = self.t2pulsar.sun_ssb

            if "ELONG" and "ELAT" in np.concatenate((t2pulsar.pars(), t2pulsar.pars(which="set"))):
                sunssb[:, :3] = utils.ecl2eq_vec(sunssb[:, :3])
                sunssb[:, 3:] = utils.ecl2eq_vec(sunssb[:, 3:])
        return sunssb

    # infrastructure for sharing Pulsar objects among processes
    # (currently Tempo2Pulsar only)
    # FIXME: why?
    # the Pulsar deflater will copy select numpy arrays to SharedMemory,
    # then replace them with pickleable objects that can be inflated
    # to numpy arrays with SharedMemory storage

    def drop_not_picklable(self):
        with contextlib.suppress(AttributeError):
            del self.t2pulsar
            logger.warning("t2pulsar object cannot be pickled and has been removed.")
        return super().drop_not_picklable()

    _todeflate = ["_designmatrix", "_planetssb", "_sunssb", "_flags"]
    _deflated = "pristine"

    def deflate(psr):  # pragma: py-lt-38
        if psr._deflated == "pristine":
            for attr in psr._todeflate:
                if isinstance(getattr(psr, attr), np.ndarray):
                    setattr(psr, attr, PulsarInflater(getattr(psr, attr)))

            psr._deflated = "deflated"

    def inflate(psr):  # pragma: py-lt-38
        if psr._deflated == "deflated":
            for attr in psr._todeflate:
                if isinstance(getattr(psr, attr), PulsarInflater):
                    setattr(psr, attr, getattr(psr, attr).inflate())

            psr._deflated = "inflated"

    def destroy(psr):  # pragma: py-lt-38
        if psr._deflated == "deflated":
            for attr in psr._todeflate:
                if isinstance(getattr(psr, attr), PulsarInflater):
                    getattr(psr, attr).destroy()

            psr._deflated = "destroyed"


# FIXME: format version?
# Current format version could be set in this file
# Reading a file with a version that might not be compatible should emit a warning
class FilePulsar(BasePulsar):
    """A Pulsar object created from the data in an HDF5 file."""

    def __init__(self, h5path, sort=True, planets=True, fmt: Optional[H5Format] = None):
        """Build a FilePulsar from an HDF5 file."""
        if fmt is None:
            fmt = derivative_file.derivative_format()
        fmt.load_from_hdf5(h5path, self)
        self._sort = sort
        self.sort_data()
        self.planets = planets


def Pulsar(*args, **kwargs):
    ephem = kwargs.get("ephem")
    clk = kwargs.get("clk")
    bipm_version = kwargs.get("bipm_version")
    planets = kwargs.get("planets", True)
    sort = kwargs.get("sort", True)
    drop_t2pulsar = kwargs.get("drop_t2pulsar", True)
    drop_pintpsr = kwargs.get("drop_pintpsr", True)
    timing_package = kwargs.get("timing_package")
    if timing_package is not None:
        timing_package = timing_package.lower()

    if pint is not None:
        toas = [x for x in args if isinstance(x, TOAs)]
        model = [x for x in args if isinstance(x, TimingModel)]

    if t2 is not None:
        t2pulsar = [x for x in args if isinstance(x, t2.tempopulsar)]

    parfile = [x for x in args if isinstance(x, str) and x.split(".")[-1] == "par"]
    timfile = [x for x in args if isinstance(x, str) and x.split(".")[-1] in ["tim", "toa"]]

    if pint and toas and model:
        return PintPulsar(toas[0], model[0], sort=sort, drop_pintpsr=drop_pintpsr, planets=planets)
    elif t2 and t2pulsar:
        return Tempo2Pulsar(t2pulsar[0], sort=sort, drop_t2pulsar=drop_t2pulsar, planets=planets)
    elif parfile and timfile:
        # Check whether the two files exist
        if not os.path.isfile(parfile[0]) or not os.path.isfile(timfile[0]):
            msg = "Cannot find parfile {0} or timfile {1}!".format(parfile[0], timfile[0])
            raise IOError(msg)

        # Obtain the directory name of the timfile, and change to it
        timfiletup = os.path.split(timfile[0])
        dirname = timfiletup[0] or "./"
        reltimfile = timfiletup[-1]
        relparfile = os.path.relpath(parfile[0], dirname)

        if timing_package is None:
            if t2 is not None:
                timing_package = "tempo2"
            elif pint is not None:
                timing_package = "pint"
            else:
                raise ValueError("No timing package available with which to load a pulsar")

        # get current directory
        cwd = os.getcwd()
        try:
            # Change directory to the base directory of the tim-file to deal with
            # INCLUDE statements in the tim-file
            os.chdir(dirname)
            if timing_package.lower == "tempo2":
                if t2 is None:
                    raise ValueError("tempo2 requested but tempo2 is not available")
                # hack to set maxobs
                maxobs = get_maxobs(reltimfile) + 100
                t2pulsar = t2.tempopulsar(relparfile, reltimfile, maxobs=maxobs, ephem=ephem, clk=clk)
                return Tempo2Pulsar(
                    t2pulsar,
                    sort=sort,
                    drop_t2pulsar=drop_t2pulsar,
                    planets=planets,
                    par_name=relparfile,
                    tim_name=reltimfile,
                )
            elif timing_package.lower() == "pint":
                if pint is None:
                    raise ValueError("PINT requested but PINT is not available")
                if (clk is not None) and (bipm_version is None):
                    bipm_version = clk.split("(")[1][:-1]
                model, toas = get_model_and_toas(
                    relparfile, reltimfile, ephem=ephem, bipm_version=bipm_version, planets=planets
                )
                os.chdir(cwd)
                return PintPulsar(toas, model, sort=sort, drop_pintpsr=drop_pintpsr, planets=planets)
            else:
                raise ValueError(f"Unknown timing package {timing_package}")
        finally:
            os.chdir(cwd)
    raise ValueError("Pulsar (par/tim) not specified in {args} or {kwargs}")



"""A file format for storing Pulsar objects in HDF5.

The key information a Pulsar object needs is the residuals
and their derivatives with respect to the parameters. This
establishes an HDF5 file to efficiently store that information
along with all other information needed to have a functioning
Pulsar object.

The key class introduced by this file is FilePulsar. A
FilePulsar object can be created from an HDF5 file, and
should then be usable interchangeably with the Pulsar
object that was used to create it.

This file also introduces a function, derivative_format,
that constructs an object describing the file format.
"""
import logging
from textwrap import dedent
from typing import List, Optional

import numpy as np
from astropy import constants as c
from astropy import units as u

from h5pulsar.h5format import H5Entry, H5Format, write_dict_to_hdf5, write_array_to_hdf5_dataset
from enterprise.pulsar import BasePulsar

logger = logging.getLogger(__name__)

format_name = "derivative_file"
format_version = "0.6.0"

# light-second unit
ls = u.def_unit("ls", c.c * 1.0 * u.s)

# DM unit (pc cm^-3)
dmu = u.def_unit("dmu", u.pc * u.cm**-3)

u.add_enabled_units([ls, dmu])

standard_introduction = dedent(
    """\
    Pulsar timing begins with a set of pulse arrival times
    and fits a model to those arrival times. The usual output
    from this process is the best-fit model parameters and
    their uncertainties, and the residuals - the difference in time or
    phase between the predicted zero phase and the observed zero phase.

    For some applications, for example searching for a
    gravitational-wave background, it is vital to include not just
    these residuals but their derivative with respect to each of the
    fit parameters. This allows construction of a linearized version
    of the timing model, which can often be analytically marginalized,
    resulting in tremendous speedups. Other applications for such
    linearized models include parameter searches in photon data.

    The purpose of this file is to provide the derivatives needed
    to construct this linear model, plus all other supporting data.
    It is stored in HDF5, a widely portable binary format that is
    extensible enough to permit project-specific information to be
    stored alongside standard values.

    This text should accompany a collection of such files in
    plain-text form, and it should also be included in all such
    files as a dataset called "README".
    """
)


def write_unit_list(h5file, name, thing, attribute):
    """Write a list of Astropy units to an HDF5 file.

    This list is in a format that Astropy can parse, based
    on the format traditionally used in FITS files.
    """
    ls = getattr(thing, attribute)
    h5file.attrs[name] = [li.to_string() for li in ls]


def write_flags(h5file, name, thing, attribute):
    """Write a set of Pulsar flags to an HDF5 file.

    These files are written as an HDF5 group whose datasets
    are the flags and their values.
    """
    value = getattr(thing, attribute)
    if isinstance(value, np.ndarray):
        # t2pulsar uses a structured dtype instead of a dictionary
        value = {flag: value[flag] for flag in value.dtype.names}
    return write_dict_to_hdf5(h5file, name, value)


def write_designmatrix(h5file, name, thing, attribute):
    """Write a design matrix to an HDF5 file.

    This writes the design matrix as an HDF5 dataset but
    also attaches attributes indicating the entry labels
    and the units.
    """
    if attribute != "_designmatrix":
        raise ValueError(f"Trying to write {attribute} as if it were the design matrix")
    new_dataset = write_array_to_hdf5_dataset(h5file, name=name, value=getattr(thing, attribute))
    if hasattr(thing, "designmatrix_units"):
        # T2Pulsar objects don't have this
        write_unit_list(new_dataset, name="units", thing=thing, attribute="designmatrix_units")
    new_dataset.attrs["labels"] = [str(f) for f in thing.fitpars]
    return new_dataset


def read_designmatrix(h5file, name, thing, attribute):
    """Read a design matrix from an HDF5 file.

    This will parse the units, if available, back into Astropy unit objects.
    """
    if attribute != "_designmatrix":
        raise ValueError(f"Trying to write {attribute} as if it were the design matrix")
    setattr(thing, attribute, np.array(h5file[name]))
    units = h5file[name].attrs.get("units")
    if units is not None:
        thing.designmatrix_units = [u.Unit(s) for s in units]


def derivative_format(
    description_intro: Optional[str] = None,
    description_finale: Optional[str] = None,
    initial_entries: Optional[List[H5Entry]] = None,
    final_entries: Optional[List[H5Entry]] = None,
) -> H5Format:
    """Construct an object describing a derivative file format.

    The returned object can be used to read and write derivative objects.
    This function allows for additional entries or description to be
    added to suit particular projects.
    """
    if description_intro is None:
        description_intro = (
            dedent(
                """\
            # Derivative information for pulsar timing

            """
            )
            + standard_introduction
            + dedent(
                """\

            ## File contents

            """
            )
        )
    if description_finale is None:
        description_finale = "\n"
    f = H5Format(
        description_intro=description_intro,
        entries=initial_entries,
        description_finale=description_finale,
        format_name=format_name,
        format_version=format_version,
    )
    f.add_entry(H5Entry(name="Name", attribute="name", use_dataset=True, description="Pulsar name."))
    f.add_entry(
        H5Entry(
            name="RAJ",
            attribute="_raj",
            use_dataset=True,
            description="Right ascension in the Julian system. In radians.",
            extra_attributes=dict(units="rad"),
        )
    )
    f.add_entry(
        H5Entry(
            name="DECJ",
            attribute="_decj",
            use_dataset=True,
            description="Declination in the Julian system. In radians.",
            extra_attributes=dict(units="rad"),
        )
    )
    f.add_entry(
        H5Entry(
            name="DM",
            attribute="_dm",
            use_dataset=True,
            description="Best-fit dispersion measure, in pc/cm^3.",
            extra_attributes=dict(units="pc/cm3"),
        )
    )
    f.add_entry(
        H5Entry(
            name="Estimated distance",
            attribute="_pdist",
            use_dataset=True,
            description="Estimated distance and uncertainty in kiloparsecs.",
            extra_attributes=dict(units="kpc"),
        )
    )
    f.add_entry(
        H5Entry(
            name="TOA integer part",
            attribute="mjdi",
            required=False,
            use_dataset=True,
            description="""\
                This is the exact TOA, converted to TDB (barycentric dynamical time)
                but not corrected for travel time in any way. In order to retain
                nanosecond accuracy, this is split into two arrays: the integer
                and the fractional parts of the MJD. This dataset contains the
                integer part.
                """,
            extra_attributes=dict(units="day"),
        )
    )
    f.add_entry(
        H5Entry(
            name="TOA fractional part",
            attribute="mjdf",
            required=False,
            use_dataset=True,
            description="""\
                This is the exact TOA, converted to TDB (barycentric dynamical time)
                but not corrected for travel time in any way. In order to retain
                nanosecond accuracy, this is split into two arrays: the integer
                and the fractional parts of the MJD. This dataset contains the
                fractional part.
                """,
            extra_attributes=dict(units="day"),
        )
    )
    f.add_entry(
        H5Entry(
            name="TOAs in seconds",
            attribute="_toas",
            use_dataset=True,
            description="""\
                Pulse time-of-arrival data, in Modified Julian Days. These
                values are barycentered, that is, converted to times
                that the pulses would have reached the solar system barycenter.
                (This depends on the pulsar sky position.) Note
                that this array has only about microsecond resolution
                and so is insufficient to do precision timing.
                """,
            extra_attributes=dict(units="s"),
        )
    )
    f.add_entry(
        H5Entry(
            name="Raw TOAs in seconds",
            attribute="_stoas",
            use_dataset=True,
            description="""\
                TOAs at the observatory; this is corrected for observatory
                clock drift but not converted to any other time system or
                adjusted to when the pulses would have reached the solar
                system barycenter. This has also been converted to *seconds*,
                that is, the Modified Julian Date has been multiplied by 86400.
                This array too has only about microsecond precision.
                """,
            extra_attributes=dict(units="s"),
        )
    )
    f.add_entry(
        H5Entry(
            name="TOA uncertainties",
            attribute="_toaerrs",
            use_dataset=True,
            description="""\
                Uncertainties on pulse time-of-arrival data (and thus on
                residuals), in seconds.
                """,
            extra_attributes=dict(units="s"),
        )
    )
    f.add_entry(
        H5Entry(
            name="Residuals",
            attribute="_residuals",
            use_dataset=True,
            description="Residuals (model minus data, in seconds).",
            extra_attributes=dict(units="s"),
        )
    )
    f.add_entry(
        H5Entry(
            name="Radio frequencies",
            attribute="_ssbfreqs",
            use_dataset=True,
            description="""\
                Radio frequency at which each TOA is observed, in MHz. This
                frequency is corrected for Doppler shift due to the
                observatory's motion around the Sun.
                """,
            extra_attributes=dict(units="MHz"),
        )
    )
    f.add_entry(
        H5Entry(
            name="Telescope names",
            attribute="_telescope",
            use_dataset=True,
            description="""\
                The name of the telescope at which each TOA was observed.
                These names are PINT- (or TEMPO2-)style telescope names (for
                example `arecibo`).
                """,
        )
    )
    f.add_entry(H5Entry(name="Fit parameters", attribute="fitpars", use_dataset=True, description="Fitted parameters."))
    f.add_entry(
        H5Entry(
            name="Design matrix",
            attribute="_designmatrix",
            use_dataset=True,
            description="""\
                Design matrix. This is an array that is (number of TOAs) by
                (number of fit parameters). Each column is the derivative of
                the residual (in seconds) with respect to the corresponding
                fit parameter. This dataset has an attribute `labels` that
                indicates the labels of the design matrix entries (which will
                be identical to the fit parameters) and `units` giving the units
                of the design matrix entries. These units are stored in Astropy's
                "generic" string format for units, which is based on that used in
                FITS files.
                """,
            write=write_designmatrix,
        )
    )
    f.add_entry(
        H5Entry(
            name="Set parameters",
            attribute="setpars",
            use_dataset=True,
            description="""\
                Parameters of the timing model that were fixed during fitting.
                Not all of these even have numeric values.
                """,
        )
    )
    f.add_entry(
        H5Entry(
            name="Par file",
            attribute="parfile",
            use_dataset=True,
            required=False,
            description="""\
                A `.par` file describing the timing model, as a string.
                This can be quite long if the model has many DMX parameters.
                The value is stored as an array of UTF-8 byte strings, one
                per line.
                """,
        )
    )
    f.add_entry(
        H5Entry(
            name="Tim file",
            attribute="timfile",
            required=False,
            use_dataset=True,
            description="""\
                A `.tim` file recording the full TOA information. This is
                in the form of an array of strings (UTF-8 encoded), one per
                line. The file is in TEMPO2 format, so will normally contain
                more lines than there are TOAs.
                """,
        )
    )
    f.add_entry(
        H5Entry(
            name="Pulsar sky position",
            attribute="_pos",
            use_dataset=True,
            description="""\
                Unit vector pointing to the pulsar's sky position, in equatorial
                coordinates.
                """,
        )
    )
    f.add_entry(
        H5Entry(
            name="Pulsar sky position as a function of time",
            attribute="_pos_t",
            use_dataset=True,
            description="""\
                Unit vector pointing to the pulsar's sky position, in equatorial
                coordinates, as a function of time (three values per TOA).
                """,
        )
    )
    f.add_entry(
        H5Entry(
            name="Sun positions",
            attribute="_sunssb",
            use_dataset=True,
            description="""\
                Sun positions (and possibly velocities) relative to
                the solar system barycenter, in light-seconds. This array
                will be (number of TOAs) by 6. If the Sun velocities are
                unavailable they will be set to zero.
                """,
            extra_attributes=dict(units="ls"),
        )
    )
    f.add_entry(
        H5Entry(
            name="Planet positions",
            attribute="_planetssb",
            use_dataset=True,
            required=False,
            description="""\
                Planet positions (and possibly velocities) relative to
                the solar system barycenter, in light-seconds. This array
                will be (number of TOAs) by 9 by 6. The planets are in order
                outward from the Sun, including Pluto. If not all planet
                positions or velocities are available, the unknown entries will
                contain NaNs. PINT generally computes only positions and only
                for the Earth, Jupiter, Saturn, Uranus, and Neptune.
                """,
            extra_attributes=dict(units="ls"),
        )
    )
    f.add_entry(
        H5Entry(
            name="DMX",
            attribute="_dmx",
            use_dataset=True,
            description="""\
                DMX information. This describes a time-variable dispersion
                measure to the pulsar using a piecewise-constant model.
                Each piece covers a specified range of TOA times and specifies
                a delta-DM that should be added to the pulsar's overall DM value
                within the corresponding time interval. This will be recorded in
                the HDF5 file as a group, with a sub-group for each DMX piece;
                the relevant values are recorded as attributes of this
                sub-group.
                """,
        )
    )
    f.add_entry(
        H5Entry(
            name="Flags",
            attribute="_flags",
            use_dataset=True,
            description="""\
                Flags associated with TOAs. The tempo2 format allows a flexible
                list of flags to be associated with each TOA; these often record
                details like the observing frontend and backend. There is a
                list of flags recommended by the International Pulsar Timing
                Array. This entry is an HDF5 group, which contains an HDF5
                dataset for each flag that occurs in the file; the dataset
                contains UTF-8-encoded string values for that flag for each TOA.
                """,
            write=write_flags,
        )
    )

    if final_entries is not None:
        for entry in final_entries:
            f.add_entry(entry)
    return f
